# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils to handle resources."""

import itertools
import os
import pathlib
import sys
import types
import typing
from typing import Union

from etils.epath import abstract_path
from etils.epath import register
from etils.epath.typing import PathLike

# pylint: disable=g-import-not-at-top
if sys.version_info >= (3, 9):  # `importlib.resources.files` was added in 3.9
  import importlib.resources as importlib_resources
else:
  import importlib_resources

if sys.version_info >= (3, 8):  # `zipfiles.Path` was added in 3.8
  import zipfile
else:
  import zipp as zipfile  # Before 3.8, importlib_resources uses backport
# pylint: enable=g-import-not-at-top


@register.register_path_cls
class ResourcePath(zipfile.Path):
  """Wrapper around `zipfile.Path` compatible with `os.PathLike`.

  Note: Calling `os.fspath` on the path will extract the file so should be
  discouraged.

  """

  def __fspath__(self) -> str:
    """Path string for `os.path.join`, `open`,...

    compatibility.

    Note: Calling `os.fspath` on the path extract the file, so should be
    discouraged. Prefer using `read_bytes`,... This only works for files,
    not directories.

    Returns:
      the extracted path string.
    """
    raise NotImplementedError('zipapp not supported. Please send us a PR.')

  if sys.version_info < (3, 10):

    # Required due to: https://bugs.python.org/issue42043
    def _next(self, at) -> 'ResourcePath':
      return type(self)(self.root, at)

    # Before 3.10, joinpath only accept a single arg
    def joinpath(self, *parts: PathLike) -> 'ResourcePath':
      """Overwrite `joinpath` to be consistent with `pathlib.Path`."""
      if not parts:
        return self
      else:
        return super().joinpath(os.path.join(*parts))  # pylint: disable=no-value-for-parameter


def resource_path(package: Union[str, types.ModuleType]) -> abstract_path.Path:
  """Returns `importlib.resources.files`."""
  path = importlib_resources.files(package)  # pytype: disable=module-attr
  if isinstance(path, pathlib.Path):
    # TODO(etils): To ensure compatibility with zipfile.Path, we should ensure
    # that the returned `pathlib.Path` isn't missused. More specifically:
    # * `os.fspath` should only be called on files (not directories)
    # * `str(path)` should be forbidden (only `__format__` allowed).
    # In practice, it is trickier to do as `__fspath__` and `__str__` are
    # called internally.
    # Convert to `GPath` for consistency and compatibility with `MockFs`.
    return abstract_path.Path(path)
  elif isinstance(path, zipfile.Path):
    path = ResourcePath(path.root, path.at)
    return typing.cast(abstract_path.Path, path)
  else:
    raise TypeError(f'Unknown resource path: {type(path)}: {path}')


def to_write_path(path: abstract_path.Path) -> abstract_path.Path:
  """Cast the path to a read-write Path."""
  return path



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python utils."""

import enum


# TODO(py3.11): Replace by `enum.StrEnum`, but keeping `Enum.__repr__`
class StrEnum(str, enum.Enum):
  """Like `Enum`, but `enum.auto()` assigns `str` rather than `int`.

  ```python
  class MyEnum(epy.StrEnum):
    SOME_ATTR = enum.auto()
    OTHER_ATTR = enum.auto()

  assert MyEnum('some_attr') is MyEnum.SOME_ATTR
  assert MyEnum.SOME_ATTR == 'some_attr'
  ```

  """

  def _generate_next_value_(name, start, count, last_values):  # pylint: disable=no-self-argument
    return name.lower()



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Array spec utils."""

from __future__ import annotations

import functools
from typing import Any, Optional

from etils import array_types
from etils.array_types import Array
from etils.enp import numpy_utils
import numpy as np

lazy = numpy_utils.lazy


class UnknownArrayError(TypeError):
  pass


@functools.lru_cache()
def _get_none_spec():  # -> tf.TypeSpec:
  """Returns the tf.NoneTensorSpec()."""
  assert lazy.has_tf
  # We need this hack as NoneTensorSpec is not exposed in the public API.
  # (see: b/191132147)
  ds = lazy.tf.data.Dataset.range(0)
  ds = ds.map(lambda x: (x, None))
  return ds.element_spec[-1]


class ArraySpec:
  """Structure containing shape/dtype."""

  __slots__ = ['shape', 'dtype']

  def __init__(self, shape, dtype):
    if numpy_utils.is_dtype_str(dtype):  # Normalize `str` dtype
      dtype = np.dtype('O')
    self.shape = tuple(shape)
    self.dtype = np.dtype(dtype)

  def __repr__(self) -> str:
    shape_str = ' '.join(['_' if s is None else str(s) for s in self.shape])
    dtype_str = array_types.typing.DTYPE_NP_TO_COMPACT_STR.get(
        self.dtype, self.dtype.name)
    return f'{dtype_str}[{shape_str}]'

  def __eq__(self, other) -> bool:
    if not isinstance(other, type(self)):
      return False
    else:
      return (other.shape, other.dtype) == (self.shape, self.dtype)

  @classmethod
  def is_array(cls, array: Any) -> bool:
    """Returns `True` if the given value can be converted to `ArraySpec`."""
    try:
      cls.from_array(array)
    except UnknownArrayError:
      return False
    else:
      return True

  @classmethod
  def from_array(cls, array: Array) -> Optional[ArraySpec]:
    """Construct the `ArraySpec` from the given array."""
    if isinstance(array, (np.ndarray, np.generic, ArraySpec)):
      shape = array.shape
      dtype = array.dtype
    elif lazy.has_jax and isinstance(
        array,
        (lazy.jax.ShapeDtypeStruct, lazy.jnp.ndarray),
    ):
      shape = array.shape
      dtype = array.dtype
    elif lazy.has_tf and isinstance(
        array,
        (lazy.tf.TensorSpec, lazy.tf.Tensor),
    ):
      shape = array.shape
      dtype = array.dtype.as_numpy_dtype
    elif lazy.has_tf and isinstance(array, type(_get_none_spec())):
      return None  # Special case for `NoneTensorSpec()`
    else:
      raise UnknownArrayError(f'Unknown array-like type: {array!r}')
    # Should we also handle `bytes` case ?
    return cls(shape=shape, dtype=dtype)



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for text_utils."""

import textwrap

from etils import epy


def test_dedent():

  text1 = textwrap.dedent("""\
      A(
         x=1,
      )""")
  text2 = epy.dedent("""
      A(
         x=1,
      )
      """)
  assert text1 == text2


def test_lines():
  d = {'a': 1, 'b': 2}

  lines = epy.Lines()
  lines += 'dict('
  with lines.indent():
    for k, v in d.items():
      lines += f'{k}={v},'
  lines += ')'
  text = lines.join()

  assert text == epy.dedent("""
      dict(
          a=1,
          b=2,
      )
      """)


def test_lines_nested_indent():

  lines = epy.Lines(indent=2)
  lines += '['
  with lines.indent():
    lines += 'a,\nb,'
    lines += '['
    with lines.indent():
      lines += 'c,\nd,'
    lines += '],'
  lines += ']'
  text = lines.join()

  assert text == epy.dedent("""
      [
        a,
        b,
        [
          c,
          d,
        ],
      ]
      """)


def test_lines_collapse():

  class A(dict):

    def __repr__(self) -> str:
      lines = epy.Lines()
      collapse = len(self) <= 1

      lines += type(self).__name__ + '('
      with lines.indent():
        for k, v in self.items():
          suffix = '' if collapse else ','
          lines += f'{k}={v}' + suffix
      lines += ')'
      return lines.join(collapse=collapse)

  assert repr(A(a=A(), b=A(a=A(a=None)))) == epy.dedent("""
      A(
          a=A(),
          b=A(a=A(a=None)),
      )
      """)

  assert repr(A(a=A(), b=A(a=A(a=None, b=A())))) == epy.dedent("""
      A(
          a=A(),
          b=A(a=A(
              a=None,
              b=A(),
          )),
      )
      """)



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unfrozen dataclasses."""

from __future__ import annotations

import dataclasses
from typing import Any, Generic, NoReturn, TypeVar, Union

from etils import epy

_Cls = TypeVar('_Cls')
_T = TypeVar('_T')


def add_unfrozen(cls: _Cls) -> _Cls:
  """Add the `frozen`, `unfrozen` methods."""
  cls_frozen = getattr(cls, 'frozen', None)
  cls_unfrozen = getattr(cls, 'has_unfrozen', None)

  # Already inherit a unfrozeen class
  if cls_frozen is frozen and cls_unfrozen is unfrozen:
    return cls

  # Partial implementation, or collision detected
  if cls_frozen is not None or cls_unfrozen is not None:
    raise ValueError(f'{cls} already define `frozen` or `unfrozen`')

  cls.frozen = frozen
  cls.unfrozen = unfrozen

  return cls


def frozen(self: _T) -> _T:
  """Freeze the dataclass."""
  raise ValueError('`.frozen()` can only be called after `.unfrozen()`.')


def unfrozen(self: _T) -> _T:
  """Returns a lazy deep-copy of the dataclass."""
  impl = _MutableProxyImpl(obj=self, common=_Common(), is_root=True)
  return impl.public_api


# Limitations: Compatibility of _MutableProxy with jax.tree_utils, chex,...


class _MutableProxy(Generic[_T]):
  """Proxy which mutate the dataclass.

  Note: To avoid attribute collisions with the wrapped class, the actual
  implementation is moved inside the `__impl` attribute.

  This module only expose the public API.

  """

  def __init__(self, impl: _MutableProxyImpl):
    super().__setattr__('_MutableProxy__impl', impl)

  def unfrozen(self) -> NoReturn:
    raise ValueError('Object is already unfrozen. Cannot call `.unfrozen()`.')

  def frozen(self) -> _T:
    return self.__impl.frozen()

  def __getattr__(self, name: str) -> Any:
    return self.__impl.getattr(name)

  def __setattr__(self, name: str, value: Any) -> None:
    return self.__impl.setattr(name, value)

  def __repr__(self) -> str:
    return f'{self.__class__.__name__}({self.__impl.obj})'


@dataclasses.dataclass
class _Common:
  """Shared variable across all nested childs of an `unfrozen()` object.

  Attributes:
    cache: Global mapping `id(_MutableProxyImpl) -> _MutableProxyImpl` to avoid
      duplicating the same object proxy  ```python a = a.unfrozen() a.x = x a.y
      = x  # `a.x` and `a.y` point to the same object a = a.frozen() assert a.x
      is a.y ```
    is_frozen: Become `True` after `.frozen()` is called. After which all
      Mutable are invalid
  """
  cache: dict[int, _MutableProxyImpl] = dataclasses.field(default_factory=dict)
  is_frozen: bool = False

  def get_proxy(self, value: Any) -> _MutableProxyImpl:
    """Returns the proxy associated with the given value, or create it."""
    id_ = id(value)
    if id_ not in self.cache:
      self.cache[id_] = _MutableProxyImpl(obj=value, common=self)
    return self.cache[id_]


@dataclasses.dataclass
class _MutableProxyImpl(Generic[_T]):
  """Proxy implementation is a separate class to avoid collisions."""
  obj: _T
  common: _Common
  is_root: bool = False

  # Child info
  attrs: dict[str, Union[_MutableProxyImpl,
                         Any]] = dataclasses.field(default_factory=dict)

  @epy.cached_property
  def public_api(self) -> _MutableProxy:
    return _MutableProxy(self)

  @epy.cached_property
  def _fields(self) -> dict[str, dataclasses.Field]:
    # Could also filter only init=True fields
    return {f.name: f for f in dataclasses.fields(self.obj)}

  def _is_dataclass_field(self, name: str) -> bool:
    """Returns True if the field is a dataclass attribute."""
    return name in self._fields  # pylint: disable=unsupported-membership-test

  def getattr(self, name: str) -> Any:
    """Returns `obj.name`."""
    if self.common.is_frozen:
      raise AttributeError('Cannot access value after the mutable was frozen.')

    # Reuse cache if it exists
    if name in self.attrs:
      value = self.attrs[name]
    else:
      value = getattr(self.obj, name)

      # Eventually wrap the value in a proxy
      if self._is_dataclass_field(name) and dataclasses.is_dataclass(value):
        value = self.common.get_proxy(value)
        self.attrs[name] = value

    if isinstance(value, _MutableProxyImpl):
      value = value.public_api
    return value

  def setattr(self, name: str, value: Any) -> None:
    """Set `obj.name`."""
    if self.common.is_frozen:
      raise AttributeError(
          'Cannot set attributes after the mutable was frozen.')

    # TODO(epot): Check that the field we're trying to overwrite is a
    # dataclass field
    if not self._is_dataclass_field(name):
      raise AttributeError(f'Cannot set {name!r}: Not a dataclass attribute.')

    # Trying to set another mutable mapping
    if isinstance(value, _MutableProxy):
      value = value._MutableProxy__impl  # pylint: disable=protected-access
      if value.common is not self.common:
        raise ValueError(
            f'Trying to mix `unfrozen` attributes. For: {name}={value}')

    # Wrapping dataclasses in proxy objects
    elif dataclasses.is_dataclass(value):
      value = self.common.get_proxy(value)

    # Storing the new value
    self.attrs[name] = value

  def frozen(self) -> _T:
    if not self.is_root:
      raise ValueError('Only the top-level dataclass can be `.frozen`')
    self.common.is_frozen = True

    return self.resolved

  @epy.cached_property  # pytype: disable=invalid-annotation
  def resolved(self) -> _T:
    """Recursivelly call `.replace` on instances which were mutated."""
    # Cached property, so that the same object is only resolved once
    new_vals = {}
    for k, v in self.attrs.items():
      if isinstance(v, _MutableProxyImpl):
        if v.obj is v.resolved:  # Skip attributes which were not mutated
          continue
        v = v.resolved
      new_vals[k] = v
    if not new_vals:  # Object wasn't mutated
      return self.obj
    return dataclasses.replace(self.obj, **new_vals)



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for resource_utils."""

import io
import os
import zipfile

from etils import epath


def test_resource():
  path = epath.resource_path('etils.epath')
  assert path.exists()
  assert path.is_dir()
  assert 'README.md' in set(p.name for p in path.iterdir())

  path = path / 'README.md'
  assert path.exists()
  assert path.read_text()


def _make_zip_file() -> zipfile.ZipFile:
  """Returns an in-memory zip file."""
  data = io.BytesIO()
  zf = zipfile.ZipFile(data, 'w')
  zf.writestr('a.txt', b'content of a')
  zf.writestr('b/c.txt', b'content of c')
  zf.writestr('b/d/e.txt', b'content of e')
  zf.writestr('b/f.txt', b'content of f')
  zf.writestr('g/h/i.txt', b'content of i')
  zf.filename = 'alpharep.zip'
  return zf


def test_resource_path():
  path = epath.resource_utils.ResourcePath(_make_zip_file())
  assert isinstance(path, os.PathLike)
  assert path.joinpath('b/c.txt').read_text() == 'content of c'
  sub_dirs = list(path.joinpath('b').iterdir())
  assert len(sub_dirs) == 3
  for p in sub_dirs:  # Childs should be `ResourcePath` instances
    assert isinstance(p, epath.resource_utils.ResourcePath)

  # Forwarded to `Path` keep the resource.
  path = epath.Path(path)
  assert isinstance(path, epath.resource_utils.ResourcePath)

  assert path.joinpath() == path
  assert path.joinpath('abc', 'def.txt').name == 'def.txt'



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Typing utils."""

from etils.array_types.typing import ArrayAliasMeta
from etils.array_types.typing import ArrayLike
import numpy as np

Array = ArrayAliasMeta(shape=None, dtype=None)
FloatArray = ArrayAliasMeta(shape=None, dtype=None)
IntArray = ArrayAliasMeta(shape=None, dtype=None)
BoolArray = ArrayAliasMeta(shape=None, dtype=np.bool_)
StrArray = ArrayAliasMeta(shape=None, dtype=np.dtype('O'))

f32 = ArrayAliasMeta(shape=None, dtype=np.float32)
ui8 = ArrayAliasMeta(shape=None, dtype=np.uint8)
ui32 = ArrayAliasMeta(shape=None, dtype=np.uint32)
i32 = ArrayAliasMeta(shape=None, dtype=np.int32)
bool_ = ArrayAliasMeta(shape=None, dtype=np.bool_)

# Random number generator jax key
PRNGKey = ui32[2]

# Keep API clean
del np



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for backports."""

import dataclasses

from etils import epy


def test_cached_property():

  @dataclasses.dataclass
  class A:
    x: int
    counter: int = 0

    @epy.cached_property
    def y(self):
      self.counter += 1
      return self.x * 10

  a = A(x=1)
  assert a.counter == 0
  assert a.y == 10  # pylint: disable=comparison-with-callable
  assert a.y == 10  # pylint: disable=comparison-with-callable
  a.x = 2  # Even after modifying x, y is still cached.
  assert a.y == 10  # pylint: disable=comparison-with-callable
  assert a.counter == 1



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for reraise_utils."""

import re

from etils import epy
import pytest


class CustomError(Exception):

  def __init__(self, *args, **kwargs):  # pylint: disable=super-init-not-called
    # Do not call super() to ensure this would work with bad code.
    self.custom_args = args
    self.custom_kwargs = kwargs


class CustomWithStrError(CustomError):

  def __str__(self) -> str:
    return f'custom error: {self.custom_args}, {self.custom_kwargs}'


def test_reraise():
  # No args
  with pytest.raises(ValueError, match='Caught: '):
    with epy.maybe_reraise(prefix='Caught: '):
      raise ValueError

  with pytest.raises(ValueError, match='Caught: With message'):
    with epy.maybe_reraise(prefix='Caught: '):
      raise ValueError('With message')

  # Custom exception
  with pytest.raises(CustomError, match='123\nCaught!') as exc_info:
    with epy.maybe_reraise(suffix='Caught!'):
      raise CustomError(123, y=345)
  assert str(exc_info.value) == '123\nCaught!'
  assert repr(exc_info.value) == "CustomError('123\\nCaught!')"
  assert exc_info.value.custom_args == (123,)
  assert exc_info.value.custom_kwargs == {'y': 345}

  # Lazy-message
  with pytest.raises(CustomError, match=re.escape('Caught: (123, {})')):
    with epy.maybe_reraise(prefix=lambda: 'Caught: '):
      raise CustomError(123, {})

  with pytest.raises(CustomError, match=re.escape("Caught: ('a', 'b', 'c')")):
    with epy.maybe_reraise(prefix='Caught: '):
      ex = CustomError(123, {})
      ex.args = 'abc'  # Not a tuple
      raise ex

  with pytest.raises(ImportError, match='Caught: With message'):
    with epy.maybe_reraise(prefix='Caught: '):
      raise ImportError('With message', name='xyz')

  with pytest.raises(FileNotFoundError, match='Caught: With message'):
    with epy.maybe_reraise(prefix='Caught: '):
      raise FileNotFoundError('With message')


def test_no_cause():
  with pytest.raises(CustomWithStrError) as exc_info:
    with epy.maybe_reraise(prefix='Caught2: '):
      with epy.maybe_reraise(prefix='Caught: '):
        raise CustomWithStrError(None, 'message', x=2)

  e = exc_info.value
  assert isinstance(e, CustomWithStrError)
  assert (
      "Caught2: Caught: custom error: (None, 'message'), {'x': 2}") == str(e)
  assert ('CustomWithStrError("Caught2: Caught: custom error: (None, '
          '\'message\'), {\'x\': 2}")') == repr(e)
  # Attributes are correctly forwarded
  assert e.custom_args == (None, 'message')
  assert e.custom_kwargs == {'x': 2}
  # Only a single cause is set (so avoid nested context)
  assert e.__cause__ is None
  assert e.__suppress_context__


def test_with_cause():
  with pytest.raises(ImportError) as exc_info:
    with epy.maybe_reraise(prefix='Caught2: '):
      with epy.maybe_reraise(prefix='Caught: '):
        try:
          e_origin = ValueError()
          raise e_origin
        except ValueError as e:
          raise ImportError from e

  e = exc_info.value
  assert 'Caught2: Caught: ' == str(e)
  assert "ImportError('Caught2: Caught: ')" == repr(e)
  # The original cause is properly forwarded (and displayed)
  assert e.__cause__ is e_origin
  assert e.__suppress_context__



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for jax3d.utils.geo_utils."""

from etils import enp
import jax.numpy as jnp
import numpy as np
import pytest
import tensorflow.experimental.numpy as tnp


@pytest.fixture(scope='module', autouse=True)
def set_tnp():
  """Enable numpy behavior."""
  # This is required to have TF follow the same casting rules as numpy
  tnp.experimental_enable_numpy_behavior(prefer_float32=True)


@pytest.mark.parametrize('xnp', [np, jnp, tnp])
def test_interp_scalar(xnp):

  vals = xnp.array([
      [-1, -1],
      [-1, 0],
      [-1, 1],
      [0.5, 1],
      [1, 1],
  ])

  #

  out = enp.interp(vals, from_=(-1, 1), to=(0, 256))
  assert isinstance(out, xnp.ndarray)

  assert xnp.allclose(
      out,
      xnp.array([
          [0, 0],
          [0, 128],
          [0, 256],
          [192, 256],
          [256, 256],
      ]),
  )
  assert xnp.allclose(
      enp.interp(vals, from_=(-1, 1), to=(0, 1)),
      xnp.array([
          [0, 0],
          [0, 0.5],
          [0, 1],
          [0.75, 1],
          [1, 1],
      ]),
  )

  vals = xnp.array([
      [255, 255, 0],
      [255, 128, 0],
      [255, 0, 128],
  ])
  assert xnp.allclose(
      enp.interp(vals, from_=(0, 255), to=(0, 1)),
      xnp.array([
          [1, 1, 0],
          [1, 128 / 255, 0],
          [1, 0, 128 / 255],
      ]),
  )
  assert xnp.allclose(
      enp.interp(vals, from_=(0, 255), to=(-1, 1)),
      xnp.array([
          [1, 1, -1],
          [1, 0.00392157, -1],
          [1, -1, 0.00392157],
      ]),
      # np upcast to float64, but jnp keep float32, so reduce precision
      atol=1e-5,
  )


@pytest.mark.parametrize('xnp', [np, jnp, tnp])
def test_interp_coords(xnp):

  coords = xnp.array([
      [-1, -1],
      [-1, 0],
      [-1, 1],
      [0.5, 1],
      [1, 1],
  ])
  assert xnp.allclose(
      enp.interp(coords, (-1, 1), (0, (1024, 256))),
      xnp.array([
          [0, 0],
          [0, 128],
          [0, 256],
          [768, 256],
          [1024, 256],
      ]))

  coords = xnp.array([
      [[0, 0], [0, 1024]],
      [[256, 256], [0, 768]],
  ])
  assert xnp.allclose(
      enp.interp(coords, (0, (256, 1024)), (0, 1)),
      xnp.array([
          [[0, 0], [0, 1]],
          [[1, 0.25], [0, 0.75]],
      ]))



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Numpy utils."""

import sys
import typing
from typing import Any, TypeVar

from etils.array_types import Array
import numpy as np

_T = TypeVar('_T')


class _LazyImporter:
  """Lazy import module.

  This allow to use `enp` with numpy only without requiring TF nor Jax.

  """

  @property
  def has_jax(self) -> bool:
    return 'jax' in sys.modules

  @property
  def has_tf(self) -> bool:
    return 'tensorflow' in sys.modules

  @property
  def jax(self):
    import jax  # pylint: disable=g-import-not-at-top  # pytype: disable=import-error
    return jax

  @property
  def jnp(self):
    import jax.numpy as jnp  # pylint: disable=g-import-not-at-top  # pytype: disable=import-error
    return jnp

  @property
  def tf(self):
    import tensorflow  # pylint: disable=g-import-not-at-top  # pytype: disable=import-error
    return tensorflow

  @property
  def tnp(self):
    import tensorflow.experimental.numpy as tnp  # pylint: disable=g-import-not-at-top  # pytype: disable=import-error
    return tnp

  def is_tf(self, x: Array) -> bool:
    return self.has_tf and isinstance(x, self.tf.Tensor)


lazy = _LazyImporter()


def get_np_module(array: Array):  # Ideally should use `-> Literal[np]:``
  """Returns the numpy module associated with the given array.

  Args:
    array: Either tf, jax or numpy array.

  Returns:
    The numpy module.
  """
  # This is inspired from NEP 37 but without the `__array_module__` magic:
  # https://numpy.org/neps/nep-0037-array-module.html
  # Note there is also an implementation of NEP 37 from the author, but look
  # overly complicated and not available at google.
  # https://github.com/seberg/numpy-dispatch
  if isinstance(array, np.ndarray):
    return np
  elif lazy.has_jax and isinstance(array, lazy.jnp.ndarray):
    return lazy.jnp
  elif lazy.has_tf and isinstance(array, lazy.tnp.ndarray):
    return lazy.tnp
  else:
    raise TypeError(
        f'Cannot infer the numpy module from array: {type(array).__name__}')


def is_dtype_str(dtype) -> bool:
  """Returns True if the dtype is `str`."""
  # tf.string.as_numpy_dtype is object
  return np.dtype(dtype).type in {np.object_, np.str_, np.bytes_}


def is_array_str(x: Any) -> bool:
  """Returns True if the given array is a `str` array.

  Note: Also returns True for scalar `str`, `bytes` values. For compatibility
  with `tensor.numpy()` which returns `bytes`

  Args:
    x: The array to test

  Returns:
    True or False
  """
  # `Tensor(shape=(), dtype=tf.string).numpy()` returns `bytes`.
  if isinstance(x, (bytes, str)):
    return True
  elif is_array(x):
    return is_dtype_str(x.dtype)
  else:
    return False


def is_array(x: Any) -> bool:
  """Returns `True` if array is np or `jnp` array."""
  if isinstance(x, np.ndarray):
    return True
  elif lazy.has_jax and isinstance(x, lazy.jnp.ndarray):
    return True
  else:
    return False


@np.vectorize
def _to_str_array(x):
  """Decodes bytes -> str array."""
  # tf.string tensors are returned as bytes, so need to convert them back to str
  return x.decode('utf8') if isinstance(x, bytes) else x


@typing.overload
def normalize_bytes2str(x: bytes) -> str:
  ...


@typing.overload
def normalize_bytes2str(x: _T) -> _T:
  ...


# Ideally could also add `BytesArray -> StrArray`, but both `bytes` and `str`
# are `StrArray`
def normalize_bytes2str(x):
  """Normalize `bytes` array to `str` (UTF-8).

  Example of usage:

  ```python
  for ex in tfds.as_numpy(ds):  # tf.data returns `tf.string` as `bytes`
    ex = tf.nest.map_structure(enp.normalize_bytes2str, ex)
  ```

  Args:
    x: Any array

  Returns:
    x: `bytes` array are decoded as `str`
  """
  if isinstance(x, str):
    return x
  if isinstance(x, bytes):
    return x.decode('utf8')
  elif is_array_str(x):
    # Note: `np.char.decode` is likely faster but don't work on `object` nor
    # bytes arrays.
    return _to_str_array(x)
  else:
    return x



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for etils.ecolab.array_as_img."""

from unittest import mock

from etils.ecolab import array_as_img
from jax import numpy as jnp
import numpy as np
import pytest
import tensorflow as tf


def test_display_array_as_image():
  with mock.patch('IPython.get_ipython', mock.MagicMock()) as ipython_mock:
    array_as_img.display_array_as_img()
  assert ipython_mock.call_count == 1


@pytest.mark.parametrize('valid_shape', [
    (28, 28),
    (28, 28, 1),
    (28, 28, 3),
])
def test_array_repr_html_valid(valid_shape):
  # 2D images are displayed as images
  assert '<img' in array_as_img._array_repr_html(jnp.zeros(valid_shape))


@pytest.mark.parametrize(
    'invalid_shape',
    [
        (7, 7),
        (28, 7),  # Only one dimension bigger than the threshold
        (28, 28, 4),  # Invalid number of dimension
        (28, 28, 0),
        (2, 28, 28),
    ],
)
def test_array_repr_html_invalid(invalid_shape):
  assert array_as_img._array_repr_html(jnp.zeros(invalid_shape)) is None


@pytest.mark.parametrize('array_cls', [
    tf.constant,
    np.array,
    jnp.array,
])
def test_array_type(array_cls):
  array = array_cls(np.zeros((50, 50)))
  assert '<img' in array_as_img._array_repr_html(array)



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Environment utils."""

import os
import sys


def is_notebook() -> bool:
  """Returns True if running in a notebook (Colab, Jupyter) environment."""
  # Use sys.module as we do not want to trigger an import (slow)
  IPython = sys.modules.get('IPython')  # pylint: disable=invalid-name
  # Check whether we're not running in a IPython terminal
  if IPython:
    get_ipython_result = IPython.get_ipython()
    if get_ipython_result and 'IPKernelApp' in get_ipython_result.config:
      return True
  return False



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tree types."""

from typing import Dict, List, Tuple, TypeVar, Union

_T = TypeVar('_T')

# TODO(b/202712189): Add `Any` to the `Union`
# Tree can be arbitrary `chex.dataclass`, jax structure,... so has to be Any
Tree = Union[_T, List['Tree'], Tuple['Tree', ...], Dict[str, 'Tree']]  # pytype: disable=not-supported-yet



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for dataclass_utils."""

import dataclasses
from typing import Any

from etils import edc
from etils import epy
import pytest


def test_kw_only():

  @edc.dataclass(kw_only=True)
  @dataclasses.dataclass(frozen=True)
  class KwOnly:
    x: Any
    y: Any

  a = KwOnly(x=1, y=2)
  assert a.x == 1
  assert a.y == 2

  with pytest.raises(TypeError, match='contructor is keyword-only.'):
    _ = KwOnly(1, 2)


@edc.dataclass
@dataclasses.dataclass(frozen=True)
class A:
  x: Any = None
  y: Any = None


@edc.dataclass
@dataclasses.dataclass(frozen=True)
class B:
  x: Any = None
  y: Any = None

  def replace(self) -> int:  # Custom replace function
    return 123


def test_replace():
  obj = object()
  x = A(y=obj)
  y = x.replace(x=123)  # pytype: disable=attribute-error
  assert x == A(y=obj)
  assert y == A(x=123, y=obj)
  assert x.y is y.y

  assert B().replace() == 123


@edc.dataclass
@dataclasses.dataclass
class R:
  x: Any = None
  y: Any = None


@edc.dataclass
@dataclasses.dataclass
class R1(R):
  z: Any = None


class R11(R1):  # Is not dataclass but `__name__` should be correct
  pass


@edc.dataclass
@dataclasses.dataclass
class R2(R):
  z: Any = None

  def __repr__(self):
    return 'R2 repr'


@edc.dataclass
@dataclasses.dataclass
class R0Field():
  pass


@edc.dataclass
@dataclasses.dataclass
class R1Field():
  x: Any = None


def test_repr():
  assert repr(R(123, R11(y='abc'))) == epy.dedent("""
  R(
      x=123,
      y=R11(
          x=None,
          y='abc',
          z=None,
      ),
  )
  """)

  # Curstom __repr__
  assert repr(R2()) == 'R2 repr'

  # When 1 or 0 field, print in a single line
  assert repr(R0Field()) == 'R0Field()'
  assert repr(R1Field()) == 'R1Field(x=None)'

  # Recursive
  x = R()
  x.x = x
  assert repr(x) == epy.dedent("""
  R(
      x=...,
      y=None,
  )
  """)



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Text utils."""

from __future__ import annotations

import contextlib
import dataclasses
import textwrap
from typing import Iterable, Iterator


@dataclasses.dataclass
class _Line:
  """Line item."""
  content: str
  indent_lvl: int
  indent_size: int


class Lines:
  """Util to build multi-line text.

  Usefull for pretty-print tools and human readable `__repr__`.

  Example:

  ```python
  d = {'a': 1, 'b': 2}

  lines = epy.Lines()
  lines += 'dict('
  with lines.indent():
    for k, v in d.items():
      lines += f'{k}={v},'
  lines += ')'
  text = lines.join()
  ```

  Output:

  ```
  dict(
      a=1,
      b=2,
  )
  ```

  """

  def __init__(self, *, indent: int = 4):
    self._lines: list[_Line] = []
    self._indent_size = indent
    self._indent_lvl = 0

  def append(self, line: str) -> None:
    """Append a new line `str`."""
    if not isinstance(line, str):
      raise TypeError(f'Lines should be added with `str`. Got {line!r}')

    self._lines.append(
        _Line(
            content=line,
            indent_lvl=self._indent_lvl,
            indent_size=self._indent_size,
        ),)

  def extend(self, iterable: Iterable[str]) -> None:
    """Append all the new line `str` from the iterable."""
    for line in iterable:
      self.append(line)

  def __iadd__(self, line: str) -> Lines:
    """Append a new line `str`."""
    self.append(line)
    return self

  @contextlib.contextmanager
  def indent(self) -> Iterator[None]:
    self._indent_lvl += 1
    try:
      yield
    finally:
      self._indent_lvl -= 1

  def join(self, *, collapse: bool = False) -> str:
    """Returns the lines.

    Args:
      collapse: If `True`, all lines are merged together in a single line.

    Returns:
      text: All lines merged together
    """
    lines = []
    for line in self._lines:
      content = line.content
      if not collapse:
        # Add the indentation to all the sub-lines
        indentation = ' ' * line.indent_lvl * line.indent_size
        content = textwrap.indent(content, indentation)
      lines.append(content)

    if collapse:
      token = ''
    else:
      token = '\n'
    return token.join(lines)


def dedent(text: str) -> str:
  r"""Wrapper around `textwrap.dedent` which also `strip()` the content.

  Before:

  ```python
  text = textwrap.dedent(
      \"\"\"\\
      A(
         x=1,
      )\"\"\"
  )
  ```

  After:

  ```python
  text = epy.dedent(
      \"\"\"
      A(
         x=1,
      )
      \"\"\"
  )
  ```

  Args:
    text: The text to dedent

  Returns:
    The dedented text
  """
  return textwrap.dedent(text).strip()



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Colab utils."""

import contextlib
import html
import io
from typing import Iterator

import IPython.display


@contextlib.contextmanager
def _collapse_std(
    *,
    name: str,
    redirect_fn,
) -> Iterator[None]:
  """Base colapsible implementation."""
  name = html.escape(name)
  f = io.StringIO()
  with redirect_fn(f):
    yield
  content = f.getvalue()
  content = html.escape(content)
  content = f'<pre><code>{content}</code></pre>'
  content = IPython.display.HTML(
      f'<details><summary>{name}</summary>{content}</details>')
  IPython.display.display(content)


@contextlib.contextmanager
def _redirect_stdall(new_target: io.StringIO) -> Iterator[None]:
  with contextlib.redirect_stderr(new_target):
    with contextlib.redirect_stdout(new_target):
      yield


@contextlib.contextmanager
def collapse(name: str = '') -> Iterator[None]:
  """Capture stderr/stdout and display it in a collapsible block.

  Args:
    name: Name of the collapsible section.

  Yields:
    None
  """
  with _collapse_std(name=name, redirect_fn=_redirect_stdall):
    yield



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tree_utils."""

import dataclasses
import sys

from etils import enp
from etils import etree as etree_lib
import numpy as np
import pytest

# pylint: disable=g-bad-import-order,g-import-not-at-top
import jax.numpy as jnp
import tensorflow as tf
# pylint: enable=g-bad-import-order,g-import-not-at-top


@pytest.fixture(
    params=[  # Run all tests 3 times (one for each backends)
        etree_lib.jax,
        etree_lib.tree,
        etree_lib.nest,
    ],)
def etree_api(request):
  yield request.param


def test_tree_parallel_map(etree_api: etree_lib.tree_utils.TreeAPI):  # pylint: disable=redefined-outer-name
  assert etree_api.parallel_map(lambda x: x * 10, {
      'a': [1, 2, 3],
      'b': [4, 5]
  }) == {
      'a': [10, 20, 30],
      'b': [40, 50]
  }


def test_tree_parallel_map_reraise(etree_api: etree_lib.tree_utils.TreeAPI):  # pylint: disable=redefined-outer-name

  def fn(x):
    raise ValueError('Bad value')

  with pytest.raises(ValueError, match='Bad value'):
    etree_api.parallel_map(fn, [1])


def test_tree_unzip(etree_api):  # pylint: disable=redefined-outer-name
  unflatten = [{'a': 1, 'b': 10}, {'a': 2, 'b': 20}, {'a': 3, 'b': 30}]
  assert list(
      etree_api.unzip({
          'a': np.array([1, 2, 3]),
          'b': np.array([10, 20, 30]),
      })) == unflatten


@dataclasses.dataclass
class Obj:
  pass


def test_spec_like(etree_api: etree_lib.tree_utils.TreeAPI):  # pylint: disable=redefined-outer-name
  obj = Obj()

  values = [
      # tf
      tf.TensorSpec((None,), dtype=tf.int32),
      # jax
      jnp.zeros((6,), dtype=np.int32),
      # np
      {
          'a': np.ones((7,), dtype=np.float32),
      },
      enp.array_spec._get_none_spec(),
      np.array(['abc', 'def']),
      # Other values are pass-through
      None,
      123,
      'abc',
      obj,
  ]

  specs = etree_api.spec_like(values)
  assert specs == [
      enp.ArraySpec((None,), dtype=np.int32),
      enp.ArraySpec((6,), dtype=np.int32),
      {
          'a': enp.ArraySpec((7,), dtype=np.float32)
      },
      None,
      enp.ArraySpec((2,), dtype=str),
      None,
      123,
      'abc',
      obj,
  ]

  out = "[i32[_], i32[6], {'a': f32[7]}, None, str[2], None, 123, 'abc', Obj()]"
  assert repr(specs) == out



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Numpy API.

When possible, utils are meant to work with
both numpy and jax.numpy.

"""

from etils.enp.array_spec import ArraySpec
from etils.enp.interp_utils import interp
from etils.enp.numpy_utils import get_np_module
from etils.enp.numpy_utils import is_array
from etils.enp.numpy_utils import is_array_str
from etils.enp.numpy_utils import is_dtype_str
from etils.enp.numpy_utils import normalize_bytes2str



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for unfrozen."""

import dataclasses
from typing import Any

import chex
from etils import edc
import pytest


@edc.dataclass(allow_unfrozen=True)
@dataclasses.dataclass(frozen=True)
class A:
  x: Any = None
  y: Any = None

  def not_a_dataclass_attr(self):
    return 123


# Only the top-level dataclass has to be decorated
@dataclasses.dataclass(frozen=True)
class B:
  x: Any = None
  y: Any = None


# Inheritance supported
@dataclasses.dataclass(frozen=True)
class C(A):
  pass


# Test with chex dataclass
@edc.dataclass(allow_unfrozen=True)
@chex.dataclass(frozen=True)
class E:
  x: Any = None


def test_unfrozen_call_twice():
  x_origin = A(y=A(x=456))

  # Can't call frozen on frozen object
  with pytest.raises(ValueError, match='can only be called after'):
    x_origin.frozen()  # pytype: disable=attribute-error

  # Can call unfrozen twice on the original object
  x = x_origin.unfrozen()  # pytype: disable=attribute-error
  x = x_origin.unfrozen()  # pytype: disable=attribute-error

  # Can't call unfrozen on unfrozen objects
  with pytest.raises(ValueError, match='Object is already unfrozen'):
    x.unfrozen()  # pytype: disable=attribute-error

  y = x.y
  x.x = 123

  # Attribute still accessible
  assert x.not_a_dataclass_attr() == 123

  # Cannot set non-existing attributes
  with pytest.raises(AttributeError, match='object has no attribute'):
    _ = x.z

  with pytest.raises(AttributeError, match='Not a dataclass attribute'):
    x.not_a_dataclass_attr = 456

  # Only top level can be freezed
  with pytest.raises(ValueError, match='Only the top-level'):
    y.frozen()

  x_freezed = x.frozen()

  # After frozen has been called, cannot use the unfrozen objects anymore
  with pytest.raises(AttributeError, match='mutable was frozen'):
    _ = x.x
  with pytest.raises(AttributeError, match='mutable was frozen'):
    x.x = 123
  with pytest.raises(AttributeError, match='mutable was frozen'):
    _ = y.x
  with pytest.raises(AttributeError, match='mutable was frozen'):
    y.x = 123

  # Cannot call unfrozen twice on the original object
  x = x_origin.unfrozen()  # pytype: disable=attribute-error

  assert x_freezed == A(x=123, y=A(x=456))
  assert x_origin == A(y=A(x=456))


def test_unfrozen_original_obj_non_mutated():
  x_origin = A(x='abc', y=A(y='def'))

  x = x_origin.unfrozen()  # pytype: disable=attribute-error
  assert x.y.y == 'def'
  x.x = 123
  x.y = 456
  assert x.x == 123
  assert x.y == 456
  x = x.frozen()

  assert x == A(x=123, y=456)

  # x origin should be not mutated
  assert x_origin == A(x='abc', y=A(y='def'))


def test_unfrozen_assigned_twice():
  x_origin = A(x='abc', y=B(y='def'))

  x = x_origin.unfrozen()  # pytype: disable=attribute-error

  x.x = x.y  # Assign the `B()` to 2 different attributes
  x.y.x = 123  # Updating one update the other
  x.x.y = 456
  x = x.frozen()
  assert x == A(x=B(x=123, y=456), y=B(x=123, y=456))
  assert x_origin == A(x='abc', y=B(y='def'))


def test_unfrozen_nested():
  x_origin = A(x=B(x=A(x=123)))
  x = x_origin.unfrozen()  # pytype: disable=attribute-error
  val = A()
  x.x.x.x = A(x=val, y=A(y=123))
  x.x.x.x.y.y = 456

  x = x.frozen()
  assert x.x.x.x.x is val
  assert x == A(x=B(x=A(x=A(x=A(), y=A(y=456)))))
  assert x_origin == A(x=B(x=A(x=123)))


def test_unfrozen_inheritance():
  x_origin = C(x=B(x=A(x=123)))
  x = x_origin.unfrozen()  # pytype: disable=attribute-error
  val = A()
  x.x.x.x = A(x=val, y=A(y=123))
  x.x.x.x.y.y = 456

  x = x.frozen()
  assert isinstance(x, C)
  assert x.x.x.x.x is val
  assert x == C(x=B(x=A(x=A(x=A(), y=A(y=456)))))
  assert x_origin == C(x=B(x=A(x=123)))


def test_unfrozen_same_object():
  a = A(x=1, y=2)
  x_origin = A(x=a, y=a)

  x = x_origin.unfrozen()  # pytype: disable=attribute-error
  x.x.x = 10
  x = x.frozen()

  # Limitation: Even if `a.x` was modified, only the member which was accessed
  # was modified (x_origin.x but not `x_origin.y`), even if they where the
  # same instance before unfrozen.
  assert x == A(x=A(x=10, y=2), y=A(x=1, y=2))
  assert x_origin == A(x=A(x=1, y=2), y=A(x=1, y=2))

  x = x_origin.unfrozen()  # pytype: disable=attribute-error
  x.x.x = 10
  x.y.y = 20
  x = x.frozen()
  # Here, the 2 members are tracked, so we can correctly update them
  assert x.x is x.y
  assert x == A(x=A(x=10, y=20), y=A(x=10, y=20))
  assert x_origin == A(x=A(x=1, y=2), y=A(x=1, y=2))


def test_unfrozen_chex():
  x_origin = E(x=E(x=A(y=E(x=123))))
  x = x_origin.unfrozen()  # pytype: disable=attribute-error
  x.x.x.y.x = 456
  x = x.frozen()
  assert isinstance(x, E)
  assert x.x.x.y.x == 456
  assert x == E(x=E(x=A(y=E(x=456))))
  assert x_origin == E(x=E(x=A(y=E(x=123))))


def test_unfrozen_no_updates():
  x_origin = A(x=A(x=A(x=123)), y=A(y=A(y=456)))

  x = x_origin.unfrozen()  # pytype: disable=attribute-error
  assert x.x.x.x == 123  # Access read-only does not trigger `.replace`
  assert x.y.y.y == 456
  x = x.frozen()

  assert x_origin.x.x.x is x.x.x.x
  assert x_origin.x.x is x.x.x
  assert x_origin.x is x.x
  assert x_origin is x
  assert x_origin.y.y.y is x.y.y.y
  assert x_origin.y.y is x.y.y
  assert x_origin.y is x.y

  x = x_origin.unfrozen()  # pytype: disable=attribute-error
  x.x.x.x = 678
  assert x.y.y.y == 456
  x = x.frozen()

  # x is modifed, but not y
  assert x_origin.x.x.x is not x.x.x.x
  assert x_origin.x.x is not x.x
  assert x_origin is not x
  assert x_origin.y.y.y is x.y.y.y
  assert x_origin.y.y is x.y.y
  assert x_origin.y is x.y


# TODO(epot): Support circles
@pytest.mark.xfail
def test_unfrozen_cicle():
  x_origin = A(y=B(y=123))
  x = x_origin.unfrozen()  # pytype: disable=attribute-error
  x.y.y = x
  x = x.frozen()
  assert x.y.y is x



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Field utils."""

from __future__ import annotations

import dataclasses
import typing
from typing import Any, Callable, Generic, Optional, Type, TypeVar

_Dataclass = Any
_In = Any
_Out = Any
_InT = TypeVar('_InT')
_OutT = TypeVar('_OutT')


def field(
    *,
    validate: Optional[Callable[[_In], _OutT]] = None,
    **kwargs: Any,
) -> dataclasses.Field[_OutT]:
  """Like `dataclasses.field`, but allow `validator`.

  Args:
    validate: A callable `(x) -> x` called each time the variable is assigned.
    **kwargs: Kwargs forwarded to `dataclasses.field`

  Returns:
    The field.
  """
  if validate is None:
    return dataclasses.field(**kwargs)
  else:
    field_ = _Field(validate=validate, field_kwargs=kwargs)
    return typing.cast(dataclasses.Field, field_)


class _Field(Generic[_InT, _OutT]):
  """Field descriptor."""

  def __init__(
      self,
      validate: Callable[[_InT], _OutT],
      field_kwargs: dict[str, Any],
  ) -> None:
    """Constructor.

    Args:
      validate: A callable called each time the variable is assigned.
      field_kwargs: Kwargs forwarded to `dataclasses.field`
    """
    # Attribute name and objtype refer to the object in which the descriptor
    # is applied. E.g. if `A.x = edc.field()`:
    # * _attribute_name = 'x'
    # * _objtype = A
    self._attribute_name: Optional[str] = None
    self._objtype: Optional[Type[_Dataclass]] = None

    self._validate = validate
    self._field_kwargs = field_kwargs

    # Whether `__get__` has not been called yet. See `__get__` for details.
    self._first_getattr_call: bool = True

  def __set_name__(self, objtype: Type[_Dataclass], name: str) -> None:
    """Bind the descriptor to the class (PEP 487)."""
    self._objtype = objtype
    self._attribute_name = name

  def __get__(
      self,
      obj: Optional[_Dataclass],
      objtype: Optional[Type[_Dataclass]] = None,
  ) -> _OutT:
    """Called when `MyDataclass.x` or `my_dataclass.x`."""
    # Called as `MyDataclass.my_attribute`
    if obj is None:
      if self._first_getattr_call:
        # Count the number of times `dataclasses.dataclass(cls)` calls
        # `getattr(cls, f.name)`.
        # The first time, we return a `dataclasses.Field` to let dataclass
        # do the magic.
        # The second time, `dataclasses.dataclass` delete the descriptor if
        # `isinstance(getattr(cls, f.name, None), Field)`. So it is very
        # important to return anything except a `dataclasses.Field`.
        # This rely on implementation detail, but seems to hold for python
        # 3.6-3.10.
        self._first_getattr_call = False
        return dataclasses.field(**self._field_kwargs)
      else:
        # TODO(epot): Could better handle default value: Either by returning
        # the default value, or raising an AttributeError. Currently, we just
        # return the descriptor:
        # assert isinstance(MyDataclass.my_attribute, _Field)
        return self
    else:
      # Called as `my_dataclass.my_path`
      return _getattr(obj, self._attribute_name)

  def __set__(self, obj: _Dataclass, value: _InT) -> None:
    """Called as `my_dataclass.x = value`."""
    # Validate the value during assignement
    _setattr(obj, self._attribute_name, self._validate(value))


# Because there is one instance of the `_Field` per class, shared across all
# class instances, we need to store the per-object state somewhere.
# The simplest is to attach the state in an extra `dict[str, value]`:
# `_dataclass_field_values`.


def _getattr(
    obj: _Dataclass,
    attribute_name: str,
) -> _Out:
  """Returns the `obj.attribute_name`."""
  _init_dataclass_state(obj)
  # Accessing the attribute before it was set (e.g. before super().__init__)
  if attribute_name not in obj._dataclass_field_values:  # pylint: disable=protected-access
    raise AttributeError(
        f"type object '{type(obj).__qualname__}' has no attribute "
        f"'{attribute_name}'")
  else:
    return obj._dataclass_field_values[attribute_name]  # pylint: disable=protected-access


def _setattr(
    obj: _Dataclass,
    attribute_name: str,
    value: _In,
) -> None:
  """Set the `obj.attribute_name = value`."""
  # Note: In `dataclasses.dataclass(frozen=True)`, obj.__setattr__ will
  # correctly raise a `FrozenInstanceError` before `DataclassField.__set__` is
  # called.
  _init_dataclass_state(obj)
  obj._dataclass_field_values[attribute_name] = value  # pylint: disable=protected-access


def _init_dataclass_state(obj: _Dataclass) -> None:
  """Initialize the object state containing all DataclassField values."""
  if not hasattr(obj, '_dataclass_field_values'):
    # Use object.__setattr__ for frozen dataclasses
    object.__setattr__(obj, '_dataclass_field_values', {})



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils to reraise."""

from __future__ import annotations

import contextlib
from typing import Callable, Iterator, NoReturn, Optional, Union

# TODO(epot):
# * Better stacktrace:
#   * Do not set `__suppress_context__` if the original exception has a
#     context (careful about nested exceptions).
#   * Do not have the reraise appear in the stacktrace ?
# * Design `epy.ReraiseMsg()` for better representation when nesting errors
#   (error in `x/y/z/`, rather than `error in x/: error in y/: error in z/:`)
# * Should automatically detect the current exception ?
# * Should unify `maybe_reraise` & `reraise` ?

# prefix/suffix can be:
# * A string
# * A lazy string (only computed if exception is reraised)
_Str = Union[str, Callable[[], str]]


def reraise(
    e: Exception,
    prefix: Optional[_Str] = None,
    suffix: Optional[_Str] = None,
) -> NoReturn:
  """Reraise an exception with an additional message.

  Benefit: Contrary to `raise ... from ...` and
  `raise Exception().with_traceback(tb)`, this function will:

  * Keep the original exception type, attributes,...
  * Avoid multi-nested `During handling of the above exception, another
    exception occurred`. Only the single original stacktrace is displayed.

  This result in cleaner and more compact error messages.

  Usage:

  ```
  try:
    fn(x)
  except Exception as e:
    epy.reraise(e, prefix=f'Error for {x}: ')
  ```

  Args:
    e: Exception to reraise
    prefix: Prefix to add to the exception message.
    suffix: Suffix to add to the exception message.
  """
  # Lazy-evaluate functions
  prefix = prefix() if callable(prefix) else prefix
  suffix = suffix() if callable(suffix) else suffix
  prefix = prefix or ''
  suffix = '\n' + suffix if suffix else ''
  msg = f'{prefix}{e}{suffix}'

  # Dynamically create an exception for:
  # * Compatibility with caller core (e.g. `except OriginalError`)

  class WrappedException(type(e)):
    """Exception proxy with additional message."""

    def __init__(self, msg):
      # We explicitly bypass super() as the `type(e).__init__` constructor
      # might have special kwargs
      Exception.__init__(self, msg)  # pylint: disable=non-parent-init-called

    def __getattr__(self, name: str):
      # Capture `e` through closure. We do not pass e through __init__
      # to bypass `Exception.__new__` magic which add `__str__` artifacts.
      return getattr(e, name)

    # The wrapped exception might have overwritten `__str__` & cie, so
    # use the base exception ones.
    __repr__ = BaseException.__repr__
    __str__ = BaseException.__str__

  WrappedException.__name__ = type(e).__name__
  WrappedException.__qualname__ = type(e).__qualname__
  WrappedException.__module__ = type(e).__module__
  new_exception = WrappedException(msg)

  # Propagate the exception:
  # * `with_traceback` will propagate the original stacktrace
  # * `from e.__cause__` will:
  #   * Propagate the original `__cause__` (likely `None`)
  #   * Set `__suppress_context__` to True, so `__context__` isn't displayed
  #     This avoid multiple `During handling of the above exception, another
  #     exception occurred:` messages when nesting `reraise`
  raise new_exception.with_traceback(e.__traceback__) from e.__cause__


@contextlib.contextmanager
def maybe_reraise(
    prefix: Optional[_Str] = None,
    suffix: Optional[_Str] = None,
) -> Iterator[None]:
  """Context manager which reraise exceptions with an additional message.

  Benefit: Contrary to `raise ... from ...` and
  `raise Exception().with_traceback(tb)`, this function will:

  * Keep the original exception type, attributes,...
  * Avoid multi-nested `During handling of the above exception, another
    exception occurred`. Only the single original stacktrace is displayed.

  This result in cleaner and more compact error messages.

  Usage:

  ```python
  with epy.maybe_reraise(prefix=f'Error for {x}:'):
    fn(x)
  ```

  Args:
    prefix: Prefix to add to the exception message. Can be a function for
      lazy-evaluation.
    suffix: Suffix to add to the exception message. Can be a function for
      lazy-evaluation.

  Yields:
    None
  """
  try:
    yield
  except Exception as e:  # pylint: disable=broad-except
    reraise(e, prefix=prefix, suffix=suffix)



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for colab/jupyter.

Usage:

```python
from etils.ecolab import array_as_img
```

"""

import base64
import io
import sys
import traceback
from typing import Any, Optional, Tuple

import IPython
import IPython.display
import numpy as np


Array = Any

_MIN_IMG_SHAPE: Tuple[int, int] = (10, 10)


def show(*objs, **kwargs) -> None:
  """Alias for `IPython.display.display`."""
  return IPython.display.display(*objs, **kwargs)


def display_array_as_img() -> None:
  """If called, 2d/3d imgage arrays will be plotted as images in colab/jupyter.

  Usage:

  >>> jax3d.utils.display_array_as_img()
  >>> np.zeros((28, 28, 3))  # Displayed as image

  """
  ipython = IPython.get_ipython()
  if ipython is None:
    return  # Non-notebook environement

  # Register the new representation fo np, tf and jax array
  print('Display big np/tf/jax arrays as image for nicer IPython display')
  formatter = ipython.display_formatter.formatters['text/html']

  # Try registering jax
  try:
    from jax import numpy as jnp  # pytype: disable=import-error  # pylint: disable=g-import-not-at-top
  except ImportError:
    pass
  else:
    # The array type is not exposed in the public API (registering jnp.ndarray
    # does not works), so dynamically extracting the type
    jax_array_cls = type(jnp.zeros(shape=()))  # DeviceArrayBase
    formatter.for_type(jax_array_cls, _array_repr_html)

  # Try registering TF
  try:
    import tensorflow as tf  # pytype: disable=import-error  # pylint: disable=g-import-not-at-top
  except ImportError:
    pass
  else:
    formatter.for_type(tf.Tensor, _array_repr_html)

  # Register np
  formatter.for_type(np.ndarray, _array_repr_html)


def _array_repr_html(array: Array) -> Optional[str]:
  """Returns the HTML `<img/>` repr, or `None` if array is not an image."""
  try:
    img = _get_image(array)
    if img is not None:
      return _html_img_repr(img)
    else:
      return None
  except Exception:
    # IPython display silence exceptions, so display it here
    traceback.print_exc()
    raise


def _html_img_repr(img: Array) -> str:
  """Generates the image and returns the HTML `<img />`."""
  from matplotlib.backends import backend_agg  # pytype: disable=import-error  # pylint: disable=g-import-not-at-top
  import matplotlib.pyplot as plt  # pytype: disable=import-error  # pylint: disable=g-import-not-at-top
  fig = plt.Figure()
  backend_agg.FigureCanvasAgg(fig)
  plt_image = fig.gca().imshow(img)
  fig.colorbar(plt_image)
  buffer = io.BytesIO()
  fig.savefig(buffer)
  bytes_value = buffer.getvalue()

  # Alternativelly could try using `IPython.display.Image`
  img_str = base64.b64encode(bytes_value).decode('ascii')  # pytype: disable=bad-return-type
  return f'<img src="data:image/png;base64,{img_str}" alt="Img" />'


def _get_image(img: Array) -> Optional[Array]:
  """Returns the normalized img, or `None` if the input is not an image."""
  # Normalize tf.Tensor into np.array
  if 'tensorflow' in sys.modules:
    import tensorflow as tf  # pytype: disable=import-error    # pylint: disable=g-import-not-at-top
    if isinstance(img, tf.Tensor):
      img = img.numpy()

  # Image should have reasonable dimensions
  if (len(img.shape) not in {2, 3} or img.shape[0] < _MIN_IMG_SHAPE[0] or
      img.shape[1] < _MIN_IMG_SHAPE[1]):
    return None

  if len(img.shape) == 2:
    return img

  assert len(img.shape) == 3
  if img.shape[-1] == 1:
    return img.squeeze(-1)
  elif img.shape[-1] == 3:
    return img
  else:  # More than 3 channels, too big to be displayed
    return None



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for typing."""

from etils.array_types import Array, f32, ui8  # pylint: disable=g-multiple-import
import jax.numpy as jnp
import pytest


def f(x: f32, y: ui8['... c']):
  return x + y


def g():
  x0: Array = jnp.array([0])
  x1: Array['...'] = jnp.array([0])
  x2: f32['...'] = jnp.array([0])
  x3: ui8['...'] = jnp.array([0])

  f(jnp.array([0]), jnp.array([0]))
  f(x0, x1)
  f(x2, x3)


@pytest.mark.parametrize(
    'alias, repr_, shape, dtype',
    [
        (Array, 'Array[...]', '...', None),  # No dtype/shape defined
        (Array[''], 'Array[]', '', None),
        (Array['h w c'], 'Array[h w c]', 'h w c', None),
        (f32, 'f32[...]', '...', jnp.float32),  # float32
        (f32['x'], 'f32[x]', 'x', jnp.float32),
        (ui8['h w'], 'ui8[h w]', 'h w', jnp.uint8),
        (f32[1], 'f32[1]', '1', jnp.float32),  # int
        (f32[()], 'f32[]', '', jnp.float32),  # tuple
        (f32[1, 3], 'f32[1 3]', '1 3', jnp.float32),  # tuple[int]
        (f32[4, 'h', 'w'], 'f32[4 h w]', '4 h w', jnp.float32),  # tuple[str]
        (f32[...], 'f32[...]', '...', jnp.float32),  # With elipsis
        (f32[..., 3], 'f32[... 3]', '... 3', jnp.float32),
    ],
)
def test_array_alias(alias, repr_, shape, dtype):
  assert repr(alias) == repr_
  assert str(alias) == repr_
  assert alias.shape == shape
  assert alias.dtype == dtype


def test_array_eq():
  assert f32['h w'] == f32['h', 'w']
  assert f32['1 2 3'] == f32[1, 2, 3]
  assert f32[...] == f32['...']

  assert f32[None] == f32['_']
  assert f32[None, 3] == f32['_ 3']
  assert f32[None, None] == f32['_ _']
  assert f32[..., None, 3] == f32['... _ 3']
  assert f32['... _', 3] == f32['... _ 3']

  assert f32['h w'] != f32['h c']
  assert f32['h w'] != Array['h w']
  assert f32['h w'] != ui8['h w']

  assert {f32['h w'], f32['h w'], f32['h', 'w']} == {f32['h w']}



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for array_spec."""

from etils import enp
import jax
import jax.numpy as jnp
import numpy as np
import tensorflow as tf


def _array_repr(x) -> str:
  return repr(enp.ArraySpec.from_array(x))


def test_array_spec_valid():
  assert enp.ArraySpec((), np.float32) == enp.ArraySpec([], np.dtype('float32'))

  assert not enp.ArraySpec.is_array(123)
  assert not enp.ArraySpec.is_array('123')
  assert enp.ArraySpec.is_array(np.array(123))
  assert enp.ArraySpec.is_array(np.array('123'))


def test_array_spec_tensors():
  # ====== Numpy ======
  # np.array
  assert _array_repr(np.zeros((3,), dtype=np.float64)) == 'f64[3]'
  assert _array_repr(np.array(123)) == 'int64[]'
  # str arrays
  assert _array_repr(np.array('123')) == 'str[]'
  assert _array_repr(np.array(['abc', 'def', ''], dtype=object)) == 'str[3]'
  assert _array_repr(np.array(['abc', 'def'])) == 'str[2]'

  # ====== Jax ======
  # jnp.array
  assert _array_repr(jnp.ones((5,), dtype=jnp.bool_)) == 'bool_[5]'
  # jax.ShapeDtypeStruct
  assert _array_repr(jax.ShapeDtypeStruct((6,), dtype=np.int32)) == 'i32[6]'

  # ====== TensorFlow ======
  # tf.Tensor
  assert _array_repr(tf.zeros((3,))) == 'f32[3]'
  # tf.TensorSpec
  assert _array_repr(tf.TensorSpec((None,), dtype=tf.int32)) == 'i32[_]'
  assert _array_repr(tf.TensorSpec((None, 3), dtype=tf.int32)) == 'i32[_ 3]'
  # str tensors
  assert _array_repr(tf.TensorSpec((4,), dtype=tf.string)) == 'str[4]'
  # tf.NoneTensorSpec
  assert enp.ArraySpec.from_array(enp.array_spec._get_none_spec()) is None

  # ====== Etils ======
  assert _array_repr(enp.ArraySpec((1, 2), dtype=np.float32)) == 'f32[1 2]'
  assert _array_repr(enp.ArraySpec((None, 2), dtype=str)) == 'str[_ 2]'
  # TODO(epot): support array_types ?


def test_array_spec_repr():
  assert repr(enp.ArraySpec((), np.float32)) == 'f32[]'
  assert repr(enp.ArraySpec((1, 3), np.uint8)) == 'ui8[1 3]'
  assert repr(enp.ArraySpec((), np.complex64)) == 'complex64[]'
  assert repr(enp.ArraySpec((4,), np.dtype('O'))) == 'str[4]'
  assert repr(enp.ArraySpec((4,), str)) == 'str[4]'
  # `str()` works too:
  assert str(enp.ArraySpec((1,), np.int32)) == 'i32[1]'



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Abstract path."""

import os
import pathlib
import typing
from typing import Any, AnyStr, Iterator, Optional, Type, TypeVar

from etils.epath.typing import PathLike

T = TypeVar('T')


# Ideally, `Path` should be `abc.ABC`. However this trigger pytype errors
# when calling `Path()` (can't instantiate abstract base class)
# Also this allow path childs to only partially implement the Path API (e.g.
# read only path)
def abstractmethod(fn: T) -> T:
  return fn


class Path(pathlib.PurePosixPath):
  """Abstract base class for pathlib.Path-like API.

  See [pathlib.Path](https://docs.python.org/3/library/pathlib.html)
  documentation.

  """

  def __new__(cls: Type[T], *args: PathLike) -> T:
    """Create a new path.

    ```python
    path = abcpath.Path()
    ```

    Args:
      *args: Paths to create

    Returns:
      path: The registered path
    """
    from etils.epath import register  # pylint: disable=g-import-not-at-top
    if cls == Path:
      if not args:
        return register.make_path('.')
      root, *parts = args
      return register.make_path(root).joinpath(*parts)
    else:
      return super().__new__(cls, *args)

  # ====== Pure paths ======

  # py3.9 backport of PurePath.is_relative_to.
  def is_relative_to(self, *other: PathLike) -> bool:
    """Return True if the path is relative to another path or False."""
    try:
      self.relative_to(*other)
      return True
    except ValueError:
      return False

  def format(self: T, *args: Any, **kwargs: Any) -> T:
    """Apply `str.format()` to the path."""
    return type(self)(os.fspath(self).format(*args, **kwargs))  # pytype: disable=not-instantiable

  # ====== Read-only methods ======

  @abstractmethod
  def exists(self) -> bool:
    """Returns True if self exists."""
    raise NotImplementedError

  @abstractmethod
  def is_dir(self) -> bool:
    """Returns True if self is a dir."""
    raise NotImplementedError

  def is_file(self) -> bool:
    """Returns True if self is a file."""
    return not self.is_dir()

  @abstractmethod
  def iterdir(self: T) -> Iterator[T]:
    """Iterates over the directory."""
    raise NotImplementedError

  @abstractmethod
  def glob(self: T, pattern: str) -> Iterator[T]:
    """Yielding all matching files (of any kind)."""
    # Might be able to implement using `iterdir` (recursivelly for `rglob`).
    raise NotImplementedError

  def rglob(self: T, pattern: str) -> Iterator[T]:
    """Yielding all matching files recursivelly (of any kind)."""
    return self.glob(f'**/{pattern}')

  def expanduser(self: T) -> T:
    """Returns a new path with expanded `~` and `~user` constructs."""
    if '~' not in self.parts:  # pytype: disable=attribute-error
      return self
    raise NotImplementedError

  @abstractmethod
  def resolve(self: T, strict: bool = False) -> T:
    """Returns the absolute path."""
    raise NotImplementedError

  @abstractmethod
  def open(
      self,
      mode: str = 'r',
      encoding: Optional[str] = None,
      errors: Optional[str] = None,
      **kwargs: Any,
  ) -> typing.IO[AnyStr]:
    """Opens the file."""
    raise NotImplementedError

  def read_bytes(self) -> bytes:
    """Reads contents of self as bytes."""
    with self.open('rb') as f:
      return f.read()

  def read_text(self, encoding: Optional[str] = None) -> str:
    """Reads contents of self as bytes."""
    with self.open('r', encoding=encoding) as f:
      return f.read()

  # ====== Write methods ======

  @abstractmethod
  def mkdir(
      self,
      mode: int = 0o777,
      parents: bool = False,
      exist_ok: bool = False,
  ) -> None:
    """Create a new directory at this given path."""
    raise NotImplementedError

  @abstractmethod
  def rmdir(self) -> None:
    """Remove the empty directory at this given path."""
    raise NotImplementedError

  @abstractmethod
  def rmtree(self) -> None:
    """Remove the directory, including all sub-files."""
    raise NotImplementedError

  @abstractmethod
  def unlink(self, missing_ok: bool = False) -> None:
    """Remove this file or symbolic link."""
    raise NotImplementedError

  def write_bytes(self, data: bytes) -> int:
    """Writes content as bytes."""
    with self.open('wb') as f:
      return f.write(data)

  def write_text(
      self,
      data: str,
      encoding: Optional[str] = None,
      errors: Optional[str] = None,
  ) -> int:
    """Writes content as str."""
    if encoding and encoding.lower() not in {'utf8', 'utf-8'}:
      raise NotImplementedError(f'Non UTF-8 encoding not supported for {self}')
    if errors:
      raise NotImplementedError(f'Error not supported for writing {self}')
    with self.open('w') as f:
      return f.write(data)

  def touch(self, mode: int = 0o666, exist_ok: bool = True) -> None:
    """Create a file at this given path."""
    if mode != 0o666:
      raise NotImplementedError(f'Only mode=0o666 supported for {self}')
    if self.exists():
      if exist_ok:
        return
      else:
        raise FileExistsError(f'{self} already exists.')
    self.write_text('')

  @abstractmethod
  def rename(self: T, target: PathLike) -> T:
    """Renames the path."""

  @abstractmethod
  def replace(self: T, target: PathLike) -> T:
    """Overwrites the destination path."""

  @abstractmethod
  def copy(self: T, dst: PathLike, overwrite: bool = False) -> T:
    """Copy the current file to the given destination."""



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interpolate utils."""

from typing import Tuple, Union

from etils.array_types import Array, ArrayLike, FloatArray  # pylint: disable=g-multiple-import
from etils.enp import numpy_utils
import numpy as np

lazy = numpy_utils.lazy

_MinMaxValue = Union[int, float, ArrayLike[Array['d']]]


def interp(
    x: Array['*d'],
    from_: Tuple[_MinMaxValue, _MinMaxValue],
    to: Tuple[_MinMaxValue, _MinMaxValue],
    axis: int = -1,
) -> FloatArray['*d']:
  """Linearly scale the given value by the given range.

  Somehow similar to `np.interp` or `scipy.interpolate.inter1d` with some
  differences like support scaling an axis by a different factors and
  extrapolate values outside the boundaries.

  `from_` and `to` are expected to be `(min, max)` tuples and the function
  interpolate between the two ranges.

  Example: Normalizing a uint8 image to `(-1, 1)`.

  ```python
  img = jnp.array([
      [0, 0],
      [127, 255],
  ])
  img = enp.interp(img, (0, 255), (0, 1))
  img == jnp.array([
      [-1, -1],
      [0.498..., 1],
  ])
  ```

  `min` and `max` can be either float values or array like structure, in which
  case the numpy broadcasting rules applies (x should be a `Array[... d]` and
  min/max values should be `Array[d]`.

  Example: Converting normalized 3d coordinates to world coordinates.

  ```python
  coords = enp.interp(coords, from_=(-1, 1), to=(0, (h, w, d)))
  ```

  * `coords[:, 0]` is interpolated from `(-1, 1)` to `(0, h)`
  * `coords[:, 1]` is interpolated from `(-1, 1)` to `(0, w)`
  * `coords[:, 2]` is interpolated from `(-1, 1)` to `(0, d)`

  Args:
    x: Array to scale
    from_: Range of x.
    to: Range to which normalize x.
    axis: Axis on which normalizing. Only relevant if `from_` or `to` items
      contains range value.

  Returns:
    Float tensor with same shape as x, but with normalized coordinates.
  """
  # Could add an `axis` argument.
  # Could add an `fill_values` argument to indicates the behavior if input
  # values are outside the input range. (`error`, `extrapolate` or `truncate`).

  # TODO(epot): Should check if `tnp.experimental_enable_numpy_behavior()`
  # is set, to check whether tf.Tensor need to be explicit casted
  # if lazy.is_tf(x) and x.dtype not in {lazy.tf.float32, lazy.tf.float64}:
  #   raise ValueError(f'`interp` input should be float32. Got: {x.dtype}')

  if axis != -1:
    raise NotImplementedError(
        'Only last axis supported for now. Please send a feature request.'
    )

  # Note: This should be static arguments so we use numpy instead of jnp
  from_ = tuple(np.array(v) for v in from_)
  to = tuple(np.array(v) for v in to)

  # `a` can be scalar or array of shape=(x.shape[-1],), same for `b`
  a, b = _linear_interp_factors(*from_, *to)  # pytype: disable=wrong-arg-types
  return  a * x + b


def _linear_interp_factors(
    old_min: _MinMaxValue,
    old_max: _MinMaxValue,
    new_min: _MinMaxValue,
    new_max: _MinMaxValue,
) -> Tuple[Union[float, FloatArray['d']], Union[float, FloatArray['d']]]:
  """Resolve the `y = a * x + b` equation and returns the factors."""
  a = (new_min - new_max) / (old_min - old_max)
  b = (old_min * new_max - new_min * old_max) / (old_min - old_max)
  return a, b



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for field_utils."""

from __future__ import annotations

import dataclasses
from typing import Any

from etils import edc
import pytest


@pytest.mark.parametrize('frozen', [True, False])
def test_field_no_op(frozen: bool):

  @dataclasses.dataclass(frozen=frozen)
  class A:
    x: Any = edc.field()  # pytype: disable=annotation-type-mismatch

  # No argument
  with pytest.raises(TypeError, match=r'__init__\(\) missing 1 required'):
    A()

  # Default argument
  a = A(123)
  assert a.x == 123

  if frozen:
    with pytest.raises(dataclasses.FrozenInstanceError):
      a.x = 456
  else:
    a.x = 456
    assert a.x == 456


@pytest.mark.parametrize('frozen', [True, False])
def test_field_multi_instance(frozen: bool):
  """Each instance has a separate state."""

  @dataclasses.dataclass(frozen=frozen)
  class A:
    x: Any = edc.field(validate=str)  # pytype: disable=annotation-type-mismatch

  a0 = A(123)
  assert a0.x == '123'
  a1 = A(456)
  assert a0.x == '123'
  assert a1.x == '456'

  if frozen:
    with pytest.raises(dataclasses.FrozenInstanceError):
      a0.x = 456
  else:
    a0.x = 789
    assert a0.x == '789'


@pytest.mark.parametrize('repr_', [True, False])
@pytest.mark.parametrize('eq', [True, False])
@pytest.mark.parametrize('order', [True, False])
@pytest.mark.parametrize('unsafe_hash', [True, False])
@pytest.mark.parametrize('frozen', [True, False])
@pytest.mark.parametrize(
    'default_kwargs',
    [{}, dict(default=789),
     dict(default_factory=lambda: 789)],
)
def test_field_validate(
    repr_: bool,
    eq: bool,
    order: bool,
    unsafe_hash: bool,
    frozen: bool,
    default_kwargs: dict[str, Any],
):
  """Make sure that default factory works for all combinations of dataclass."""
  if order and not eq:  # eq must be true if order is true
    return

  @dataclasses.dataclass(
      # default_factory doesn't make sense if init = False
      init=True,
      repr=repr_,
      eq=eq,
      order=order,
      unsafe_hash=unsafe_hash,
      frozen=frozen,  # pytype: disable=not-supported-yet
  )
  class A:
    x: Any = edc.field(validate=str, **default_kwargs)  # pytype: disable=annotation-type-mismatch

  # No argument
  if default_kwargs:
    a = A()
    assert a.x == '789'
  else:
    # No argument
    with pytest.raises(TypeError, match=r'__init__\(\) missing 1 required'):
      A()

  # Single argument
  a = A(456)
  assert a.x == '456'

  if repr_:
    assert repr(a) == "test_field_validate.<locals>.A(x='456')"

  # Updating the value should only work for non-frozen instances.
  if frozen:
    with pytest.raises(dataclasses.FrozenInstanceError):
      a.x = 678
  else:
    a.x = 678
    assert a.x == '678'

  # Class attribute
  assert isinstance(A.x, edc.field_utils._Field)



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""GPath wrapper around the gfile API."""

import ntpath
import os
import pathlib
import posixpath
import types
import typing
from typing import Any, ClassVar, Iterator, Optional, Type, TypeVar, Union

from etils import epy
from etils.epath import abstract_path
from etils.epath.typing import PathLike
import tensorflow as tf

_P = TypeVar('_P')

URI_PREFIXES = ('gs://', 's3://')
_URI_SCHEMES = frozenset(('gs', 's3'))

_URI_MAP_ROOT = {
    'gs://': '/gs/',
    's3://': '/s3/',
}


class _GPath(abstract_path.Path):
  """Pathlib like api around `tf.io.gfile`."""

  # `_PATH` is `posixpath` or `ntpath`.
  # Use explicit `join()` rather than `super().joinpath()` to avoid infinite
  # recursion.
  # Do not use `os.path`, so `PosixGPath('gs://abc')` works on windows.
  _PATH: ClassVar[types.ModuleType]

  def __new__(cls: Type[_P], *parts: PathLike) -> _P:
    full_path = '/'.join(os.fspath(p) for p in parts)
    if full_path.startswith(URI_PREFIXES):
      prefix = full_path[:5]
      new_prefix = _URI_MAP_ROOT[prefix]
      return super().__new__(cls, full_path.replace(prefix, new_prefix, 1))
    else:
      return super().__new__(cls, *parts)

  def _new(self: _P, *parts: PathLike) -> _P:
    """Create a new `Path` child of same type."""
    return type(self)(*parts)

  # Could try to use `cached_property` when beam is compatible (currently
  # raise mutable input error).
  @property
  def _uri_scheme(self) -> Optional[str]:
    if (len(self.parts) >= 2 and self.parts[0] == '/' and
        self.parts[1] in _URI_SCHEMES):
      return self.parts[1]
    else:
      return None

  @property
  def _path_str(self) -> str:
    """Returns the `__fspath__` string representation."""
    uri_scheme = self._uri_scheme
    if uri_scheme:  # pylint: disable=using-constant-test
      return self._PATH.join(f'{uri_scheme}://', *self.parts[2:])
    else:
      return self._PATH.join(*self.parts) if self.parts else '.'

  def __fspath__(self) -> str:
    return self._path_str

  def __str__(self) -> str:  # pylint: disable=invalid-str-returned
    return self._path_str

  def __repr__(self) -> str:
    return f'{type(self).__name__}({self._path_str!r})'

  def exists(self) -> bool:
    """Returns True if self exists."""
    return tf.io.gfile.exists(self._path_str)

  def is_dir(self) -> bool:
    """Returns True if self is a directory."""
    return tf.io.gfile.isdir(self._path_str)

  def iterdir(self: _P) -> Iterator[_P]:
    """Iterates over the directory."""
    for f in tf.io.gfile.listdir(self._path_str):
      yield self._new(self, f)

  def expanduser(self: _P) -> _P:
    """Returns a new path with expanded `~` and `~user` constructs."""
    return self._new(self._PATH.expanduser(self._path_str))

  def resolve(self: _P, strict: bool = False) -> _P:
    """Returns the abolute path."""
    return self._new(self._PATH.abspath(self._path_str))

  def glob(self: _P, pattern: str) -> Iterator[_P]:
    """Yielding all matching files (of any kind)."""
    for f in tf.io.gfile.glob(self._PATH.join(self._path_str, pattern)):
      yield self._new(f)

  def mkdir(
      self,
      mode: int = 0o777,
      parents: bool = False,
      exist_ok: bool = False,
  ) -> None:
    """Create a new directory at this given path."""
    if self.exists() and not exist_ok:
      raise FileExistsError(f'{self._path_str} already exists.')

    if parents:
      tf.io.gfile.makedirs(self._path_str)
    else:
      tf.io.gfile.mkdir(self._path_str)

  def rmdir(self) -> None:
    """Remove the empty directory."""
    if not self.is_dir():
      raise NotADirectoryError(f'{self._path_str} is not a directory.')
    if list(self.iterdir()):
      raise ValueError(f'Directory {self._path_str} is not empty')
    tf.io.gfile.rmtree(self._path_str)

  def rmtree(self) -> None:
    """Remove the directory."""
    tf.io.gfile.rmtree(self._path_str)

  def unlink(self, missing_ok: bool = False) -> None:
    """Remove this file or symbolic link."""
    try:
      tf.io.gfile.remove(self._path_str)
    except tf.errors.NotFoundError as e:
      if missing_ok:
        pass
      else:
        raise FileNotFoundError(str(e))

  def open(
      self,
      mode: str = 'r',
      encoding: Optional[str] = None,
      errors: Optional[str] = None,
      **kwargs: Any,
  ) -> typing.IO[Union[str, bytes]]:
    """Opens the file."""
    if errors:
      raise NotImplementedError
    if encoding and not encoding.lower().startswith(('utf8', 'utf-8')):
      raise ValueError(f'Only UTF-8 encoding supported. Not: {encoding}')
    gfile = tf.io.gfile.GFile(self._path_str, mode, **kwargs)
    gfile = typing.cast(typing.IO[Union[str, bytes]], gfile)
    return gfile

  def rename(self: _P, target: PathLike) -> _P:
    """Rename file or directory to the given target."""
    # Note: Issue if WindowsPath and target is gs://. Rather than using `_new`,
    # `GPath.__new__` should dynamically return either `PosixGPath` or
    # `WindowsPath`, similarly to `pathlib.Path`.
    target = self._new(target)
    tf.io.gfile.rename(self._path_str, os.fspath(target))
    return target

  def replace(self: _P, target: PathLike) -> _P:
    """Replace file or directory to the given target."""
    target = self._new(target)
    tf.io.gfile.rename(self._path_str, os.fspath(target), overwrite=True)
    return target

  def copy(self: _P, dst: PathLike, overwrite: bool = False) -> _P:
    """Remove the directory."""
    # Could add a recursive=True mode
    dst = self._new(dst)
    tf.io.gfile.copy(self._path_str, os.fspath(dst), overwrite=overwrite)
    return dst


class PosixGPath(_GPath):
  """Pathlib like api around `tf.io.gfile`."""
  _PATH = posixpath


class WindowsGPath(pathlib.PureWindowsPath, _GPath):
  """Pathlib like api around `tf.io.gfile`."""
  _PATH = ntpath



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for numpy_utils."""

from etils import enp
import jax
import jax.numpy as jnp
import numpy as np
import pytest
import tensorflow as tf
import tensorflow.experimental.numpy as tnp


@pytest.fixture(scope='module', autouse=True)
def setup_tf():
  tnp.experimental_enable_numpy_behavior()


def fn(x):
  xnp = enp.get_np_module(x)
  y = xnp.sum(x) + x.mean()
  return x + y


def test_get_array_module_tf():
  y = fn(tf.constant([123]))
  assert isinstance(y, tnp.ndarray)


def test_lazy():
  lazy = enp.numpy_utils.lazy

  assert lazy.has_tf
  assert lazy.has_jax

  assert lazy.tf is tf
  assert lazy.jax is jax
  assert lazy.jnp is jnp

  assert lazy.is_tf(tf.constant([123]))
  assert not lazy.is_tf(np.array([123]))


@pytest.mark.parametrize('xnp', [np, jnp, tnp])
def test_get_array_module(xnp):
  y = fn(xnp.array([123]))
  assert isinstance(y, xnp.ndarray)


@pytest.mark.parametrize('xnp', [np, jnp])
def test_not_array_str(xnp):
  x = xnp.array([123])
  assert enp.is_array(x)
  assert not enp.is_array_str(x)
  assert not enp.is_dtype_str(x.dtype)


_STR_DTYPES = [
    np.dtype('<U3'),
    np.dtype('<S3'),
    np.str_,
    np.bytes_,
    str,
    bytes,
    object,
]

_STR_ARRAYS = [
    np.array(['abc', 'def']),
    np.array([b'abc', b'def']),
    np.array(['abc', 'def'], dtype=object),
    np.array([b'abc', b'def'], dtype=object),
]


@pytest.mark.parametrize('array', _STR_ARRAYS)
def test_array_str(array):
  assert enp.is_array(array)
  assert enp.is_array_str(array)
  assert enp.is_dtype_str(array.dtype)


def test_array_str_scalar():
  assert enp.is_array_str('abc')
  assert enp.is_array_str(b'abc')


@pytest.mark.parametrize('dtype', _STR_DTYPES)
def test_is_dtype_str(dtype):
  assert enp.is_dtype_str(dtype)


@pytest.mark.parametrize('dtype', [
    np.dtype(int),
    np.int64,
    int,
])
def test_is_not_dtype_str(dtype):
  assert not enp.is_dtype_str(dtype)


@pytest.mark.parametrize('array', _STR_ARRAYS)
def test_normalize_bytes2str(array):
  assert np.array_equal(
      enp.normalize_bytes2str(array),
      np.array(['abc', 'def']),
  )


def test_normalize_bytes2str_static():
  assert enp.normalize_bytes2str('abc') == 'abc'
  assert enp.normalize_bytes2str(b'abc') == 'abc'
  assert isinstance(enp.normalize_bytes2str('abc'), str)
  assert isinstance(enp.normalize_bytes2str(b'abc'), str)
  assert enp.normalize_bytes2str(123) == 123

  assert np.array_equal(
      enp.normalize_bytes2str(np.array([123, 456])),
      np.array([123, 456]),
  )
  assert isinstance(enp.normalize_bytes2str(jnp.array([1, 2, 3])), jnp.ndarray)



# Copyright 2022 The etils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tree API backends."""

from __future__ import annotations

import abc
import types
from typing import Any, Callable, TypeVar

from etils import epy
from etils.etree.typing import Tree

_T = TypeVar('_T')
_Tin = Any  # TypeVar('_Tin')
_Tout = Any  # TypeVar('_Tout')

# Structure which allow to reconstruct the tree
# * jax: TreeDef
# * tf.nest/tree: The original Tree
_TreeDef = Any


class Backend(abc.ABC):
  """Tree API backend.

  Note: The backend lazy-import the module on first call. This
  allow to use `etree` with Jax even if TF isn't installed (and
  vice-versa).
  """

  @epy.cached_property
  def module(self) -> types.ModuleType:
    """Module used by the backend."""
    try:
      module = self.import_module()
    except ImportError as e:
      epy.reraise(e, suffix=f'etree backend require {self.MODULE_NAME!r}.')
    return module

  @abc.abstractmethod
  def import_module(self) -> types.ModuleType:
    """Import and return the module."""
    raise NotImplementedError

  @abc.abstractmethod
  def map(
      self,
      map_fn: Callable[..., _Tout],  # Callable[[_Tin0, _Tin1,...], Tout]
      *trees: Tree[_Tin],  # _Tin0, _Tin1,...
  ) -> Tree[_Tout]:
    """Like `tf.nest.map_structure`."""
    raise NotImplementedError

  @abc.abstractmethod
  def flatten(self, tree: Tree[_T]) -> tuple[list[_T], _TreeDef]:
    """Like `tf.nest.flatten`."""
    raise NotImplementedError

  @abc.abstractmethod
  def unflatten(self, structure: _TreeDef, flat_sequence: list[_T]) -> Tree[_T]:
    raise NotImplementedError


class Jax(Backend):
  """`jax.tree_util` backend."""

  def import_module(self):
    import jax  # pylint: disable=g-import-not-at-top  # pytype: disable=import-error
    return jax.tree_util

  def map(self, map_fn, *trees):
    return self.module.tree_multimap(map_fn, *trees)

  def flatten(self, tree):
    flat_vals, treedef = self.module.tree_flatten(tree)
    return flat_vals, treedef

  def unflatten(self, structure, flat_sequence):
    return structure.unflatten(flat_sequence)


class DmTree(Backend):
  """`tree` backend."""

  def import_module(self):
    import tree  # pylint: disable=g-import-not-at-top  # pytype: disable=import-error
    return tree

  def map(self, map_fn, *trees):
    return self.module.map_structure(map_fn, *trees)

  def flatten(self, tree):
    return self.module.flatten(tree), tree

  def unflatten(self, structure, flat_sequence):
    return self.module.unflatten_as(structure, flat_sequence)


class Nest(Backend):
  """`tf.nest` backend."""

  def import_module(self):
    import tensorflow as tf  # pylint: disable=g-import-not-at-top  # pytype: disable=import-error
    return tf.nest

  def map(self, map_fn, *trees):
    return self.module.map_structure(map_fn, *trees)

  def flatten(self, tree):
    return self.module.flatten(tree), tree

  def unflatten(self, structure, flat_sequence):
    return self.module.pack_sequence_as(structure, flat_sequence)
